#AUTOGENERATED! DO NOT EDIT! File to edit: dev/00_utils.ipynb (unless otherwise specified).

__all__ = ['IMG_EXT', 'cd_root', 'extension_counts', 'stop_instance']

#Cell
import boto3
from collections import Counter
import mimetypes
import os
import requests

#Cell
IMG_EXT = {k[1:] for k, v in mimetypes.types_map.items() if 'image' in v}

#Cell
def cd_root(root_subdir='pylib'):
    """Run at start of Jupyter notebook to enter project root.

    Parameters
    ----------
    root_subdir: str
        Name of a subdirectory contained in the project root directory.
        If not found in the current working directory, this will move
        to the parent directory.

    Examples
    --------
    Sample file structure (abbreviated):

    delphi/
        pylib/
            setup.py
            delphi/
                __init__.py
                utils.py
        py/
            fetch_raw_data.py
        notebooks/
            nb01_eda.ipynb

    Running cd_root() from nb01_eda.ipynb will change the working
    directory from notebooks/ to delphi, which is typically the
    same directory we'd run scripts in py/ from. This makes converting
    from notebooks to scripts easier.
    """
    if root_subdir not in next(os.walk('.'))[1]:
        os.chdir('..')
    print('Current directory:', os.getcwd())

#Cell
def extension_counts(paths, common_only=True, exts=None):
    """Given a sequence of paths, find the number of times each extension
    occurs.

    Parameters
    ----------
    paths: list[str]
        Sequence of paths that can be iterated over.
    common_only: bool
        Specifies whether to limit counts to include only certain extensions,
        either passed in by the user or pulled from the mimetypes library.
    exts: set[str]
        Set of file extensions to include (format should not start with '.').
        If common_only is True and exts is None, a set of common file
        extensions are pulled from the mimetypes library.

    Returns
    -------
    dict[str, int]: dict-like object mapping file extension to
        the number of times it appears in `paths`. Files with no extension
        are represented by ''.
    """
    exts = exts or set(ext.lstrip('.') for ext in mimetypes.types_map.keys())
    counts = Counter(os.path.splitext(path)[-1].lstrip('.') for path in paths)
    if common_only:
        counts = {k: v for k, v in counts.items() if k in exts}
    return dict(counts)

#Cell
def stop_instance(debug=False, timeout=5):
    """Stop an active ec2 instance.

    Parameters
    ----------
    debug: bool
        If True, run the stop command in dry run mode. This checks if the
        user has required permissions to execute the command.
    timeout: int
        Max amount of time for request to get ec2 information. If we
        are not on an EC2, this would hang indefinitely otherwise.

    Returns
    -------
    None
    """
    url = 'http://169.254.169.254/latest/dynamic/instance-identity/document'
    try:
        r = requests.get(url, timeout=timeout).json()
    except requests.ReadTimeout as e:
        print('Request timed out. Failed to shutdown instance.')
        return
    id_, region = r['instanceId'], r['region']
    ec2 = boto3.client('ec2', region_name=region)
    ec2.stop_instances(InstanceIds=[id_], DryRun=debug)


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/01_experiments.ipynb (unless otherwise specified).

__all__ = ['snapshot_protobufs', 'fetch_experiments_data']

#Cell
from functools import partial
import os
import pandas as pd

from accio import db
from accio.prototool import ProtoTool

#Cell
def snapshot_protobufs(snapshot, limit=None, verbose=True):
    """Find samples obtained via the experiments UI from a specified snapshot.

    Parameters
    -----------
    snapshot: int
        Number of snapshot in experiments portal.
    verbose: bool
        If True, print the number of samples retrieved.

    Returns
    --------
    dict
        Keys are s3 paths, values are dictionaries containing parsed protobuf
        attributes as well as data from experiment_snapshot_sites table.
    """
    q = f"""select s3_bucket_location, url, id, experiment_snapshot_id,
        account_id, sub_account_id, org_id, ou, seq_id, created_at
        from experiment_snapshot_sites
        where experiment_snapshot_id = {snapshot}
        """
    if limit: q += f'limit {limit}'

    # Get df w/ 1 row for each page in the sample.
    rows = db.query(q, 'experiments-prod', 'reader')
    if verbose:
        print(f'{rows.shape[0]} samples retrieved.')

    # Map s3 path to dict with URL and ID, then retrieve protobuf files.
    path2data = {f's3://{row[0]}/scraped_dom.proto': dict(row[1:])
                 for i, row in rows.iterrows()}
    output = ProtoTool().buf_to_dict(path2data.keys())

    # Add URLs and IDs to feature dicts.
    return {path: {**buf, **path2data[path]} for path, buf in output.items()}

#Cell
def fetch_experiments_data(*snapshot_ids, df=True, s3_paths=False, limit=None):
    """
    Parameters
    -----------
    snapshot_ids: int (1 or more)
        Integer IDs corresponding to the relevant experiment snapshots.
        These can be found in the experiments V1 portal on the `Snapshots`
        page.
    df: bool (default True)
        If True, return the data as a single Pandas DataFrame.
    s3_paths: bool (default False)
        If True, include the urls for the protobufs in S3 in the output
        dataframe or dict. When returning a dictionary (i.e. df=False),
        s3_paths will always be present as keys.
    limit: int or None
        If int, sets the max combined number of rows to retrieve from db.
        If None, retrieve all rows. Note: if using n snapshot
        id's, each snapshot will have at most limit//n rows.

    Returns
    --------
    pd.DataFrame
        1 row for each example collected from snapshot.

    or

    dict[str, dict]
        Maps S3 file location to dict containing file attributes and site
        metadata.
    """
    if limit: limit //= len(snapshot_ids)
    proto_rounds = (snapshot_protobufs(i, limit) for i in snapshot_ids)
    data = {k: v for group in proto_rounds for k, v in group.items()}
    if df:
        if s3_paths:
            data = pd.DataFrame(data).T.reset_index()\
                     .rename({'index': 's3'}, axis=1)
        else:
            data = pd.DataFrame(list(data.values()))
    return data


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/05_cli.ipynb (unless otherwise specified).

__all__ = ['ClickLiteralOption']

#Cell
from ast import literal_eval
import click
from functools import partial

#Cell
class ClickLiteralOption(click.Option):
    """Command line option that lets user input Python data structures
    which aren't offered by Click.

    Examples
    --------
    # Notice that the input structure must be enclosed in quotes.

    @click.command(help='Sample command.')
    @click.option('--lr', cls=ClickLiteralOption, default="[3e-3, 3e-4, 1e-5]")
    def train(lr):
        # Code to train model.
    """

    def type_cast_value(self, ctx, val):
        try:
            return literal_eval(val)
        except:
            raise click.BadParameter('Failed to parse arg as Python literal.')

#Cell
click.command = partial(
    click.command,
    context_settings={'help_option_names': ['--help', '-h']}
)
click.option = partial(click.option, show_default=True)


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/03_ml.ipynb (unless otherwise specified).

__all__ = ['Debugger', 'build_pipeline', 'parse_gs_results']

#Cell
import os
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import make_column_transformer
from sklearn.impute import SimpleImputer
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import OneHotEncoder

from htools import save

#Cell
class Debugger(BaseEstimator, TransformerMixin):
    """Transformer to be used to check the output of sklearn pipelines."""

    def __init__(self, verbose=True, path=None):
        """
        Parameters
        -----------
        verbose: bool
            If True, print the shape and null count of the transformed feature
            array.
        path: str (Optional)
            If specified, save the processed data to a pickle file. If left as
            None, output will not be saved. Must use pkl file extension.
        """
        self.verbose = verbose
        self.path = path
        if path: os.makedirs(os.path.dirname(path), exist_ok=True)

    def transform(self, X, y=None):
        if self.verbose:
            print(f'X shape: {X.shape}')
            print(f'# of nulls: {np.isnan(X).sum()}')
            print(pd.DataFrame(X[:5]))
        if self.path: save(X, self.path)
        return X

    def fit(self, X, y=None):
        return self

#Cell
def build_pipeline(x, bin_cols=None, num_cols=None, cat_cols=None,
                    scaler=None, num_impute='median'):
    """Wrapper to build scikit-learn pipelines that impute missing values,
    scale numeric columns, and one hot encode categorical variables. A column
    transformer is returned. No classifier is passed in, so fit and transform
    are available methods but predict is not.

    Parameters
    -----------
    x: pd.DataFrame
        (Optional) A dataframe of features. Lists of numeric and categorical
        features will be generated automatically based on dtypes.
    scaler: sklearn.preprocessing.data
        A scikit-learn scaler object (already instantiated). If none is
        provided, mean normalization will be used.
    num_impute: str
        Strategy to use when imputing missing values in numeric (continuous)
        features. One of ('mean', 'median', 'most_frequent'), default 'median'.
        Note that categorical features will be automatically imputed with
        'most_frequent'.

    Returns
    --------
    sklearn.compose._column_transformer.ColumnTransformer
    """
    scaler = scaler or StandardScaler()

    # Dummy cols need no additional processing.
    bin_cols = bin_cols or [col for col in x.columns
                            if col.startswith('contains_')
                            or col.startswith('is_')]
    num_cols = num_cols or [col for col in x.select_dtypes(np.number).columns
                            if col not in bin_cols]
    cat_cols = cat_cols or x.select_dtypes('category').columns.tolist()

    # Process continuous and categorical variables separately.
    bin_pipe = make_pipeline(SimpleImputer(strategy='most_frequent'))
    num_pipe = make_pipeline(SimpleImputer(strategy=num_impute),
                             scaler)
    cat_pipe = make_pipeline(SimpleImputer(strategy='most_frequent'),
                             OneHotEncoder(drop='first'))

    # Combine into single transformer and add classifier to end of pipeline.
    return make_column_transformer((bin_pipe, bin_cols),
                                   (num_pipe, num_cols),
                                   (cat_pipe, cat_cols),
                                   remainder='passthrough')

#Cell
def parse_gs_results(results, metric, sort_param=None):
    """Parse dictionary returned by grid search cv_results_ so we can examine
    how metrics varied depending on hyperparameter choices.

    Parameters
    -----------
    results: dict
        cv_results_ dict.
    sort_param: str
        Hyperparameter to sort by.
    metric: str

    Returns
    --------
    pd.DataFrame
    """
    sort_by = [metric]
    if sort_param:
        sort_by.insert(0, sort_param)

    return pd.concat([pd.DataFrame(data['params']),
                      pd.DataFrame(data[f'mean_test_{metric}'],
                                   columns=[metric])],
                     axis=1).sort_values(sort_by, ascending=False)


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/07_scraping.ipynb (unless otherwise specified).

__all__ = ['fetch_img_from_url', 'load_praw_credentials', 'is_imgur_page', 'fetch_imgur_url',
           'scrape_subreddit_images_1k', 'scrape_subreddit_images', 'remove_corrupt_images']

#Cell
from bs4 import BeautifulSoup
import cv2
import glob
import logging
import os
import pandas as pd
from pathlib import Path
from PIL import Image
import praw
import re
import requests

from accio.s3tool import S3tool
from .utils import IMG_EXT

#Cell
def fetch_img_from_url(url, out_path, verbose=False):
    """Given a URL, fetch an image and download it to the specified path.

    Parameters
    ----------
    url: str
        Location of image online.
    out_path: str
        Path to download the image to.
    verbose: bool
        If True, prints a message alerting the user when the image could not be
        retrieved.

    Returns
    -------
    bool: Specifies whether image was successfully retrieved.
    """
    try:
        with requests.get(url, stream=True, timeout=10) as r:
            if r.status_code != 200:
                if verbose: print(f'STATUS CODE ERROR: {url}')
                return False

            # Write bytes to file chunk by chunk.
            with open(out_path, 'wb') as f:
                for chunk in r.iter_content(256):
                    f.write(chunk)

    # Any time url cannot be accessed, don't care about exact error.
    except Exception as e:
        if verbose: print(e)
        return False

    return True

#Cell
def load_praw_credentials(bucket='gg-datascience',
                          s3_path='hmamin/praw_creds.csv'):
    """Load credentials for dune script owned by hdmamin user on reddit.

    Returns
    -------
    dict[str, str]: Keyword arguments that can be passed directly to
        praw.Reddit to instantiate a new instance.
    """
    tmp_dir = '/tmp'
    tmp_file = os.path.join(tmp_dir, 'praw_creds.csv')
    s3 = S3tool()
    s3.download_file(bucket, s3_path, tmp_dir, maintain_s3_partitions=False)

    # Read credentials into dict and delete temporary file.
    with open(tmp_file, 'r') as f:
        cols, vals = [line.strip().split(',') for line in f]
    os.remove(tmp_file)
    return dict(zip(cols, vals))

#Cell
def is_imgur_page(url):
    """Check if a URL is a link to an imgur html page. Use as validation before
    trying to fetch an image directly.

    Returns
    -------
    bool
    """
    url = url.lower()
    return 'imgur' in url and not any(ext in url.split('.')[-1]
                                      for ext in IMG_EXT)

#Cell
def fetch_imgur_url(url):
    """Fetch and extract an image url from an imgur page. Returns None if an
    image is not found.
    """
    try:
        r = requests.get(url)

    # If url throws error, see if it is one of the links formatted like this:
    # ex: http://[Imgur](http://i.imgur.com/1gmLlya)
    # If so, recursive call on part in parens. Otherwise, return None.
    except:
        parens_links = re.findall(r'\((.+)\)', url)
        if parens_links:
            return fetch_imgur_url(parens_links[0])
        else:
            return

    if r.status_code != 200: return
    img_links = BeautifulSoup(r.content, features='lxml')\
        .find_all('link', rel='image_src')
    try:
        return img_links[0]['href']
    except:
        return

#Cell
def scrape_subreddit_images_1k(sub_name, out_dir, n=None, verbose=True,
                               print_freq=50):
    """Scrape subreddit using PRAW. Limited to 1,000 posts per call, and
    haven't found way to use this iteratively to get all posts.

    Parameters
    ----------
    sub_name: str
        Name of subreddit.
    out_dir: str
        Name of output image directory.
    n: int
        Max number of posts to scrape. If None, scrapes as many as the
        subreddit has (or until Praw limit is reached).
    verbose: bool

    print_freq: n
        Frequency with which to print message with number of posts scraped.
        Makes monitoring easier.

    Returns
    -------
    pd.DataFrame: Two columns consisting of image urls and post urls.
    """
    img_dir = os.path.join(out_dir, 'test', sub_name)
    os.makedirs(img_dir, exist_ok=True)
    creds = load_praw_credentials()
    reddit = praw.Reddit(**creds)
    sub = reddit.subreddit(sub_name)

    # Scrape top level image from each post.
    posts = []
    for i, post in enumerate(sub.top(limit=n), 1):
        if i % print_freq == 0: print(f'Scraping post {i}...')
        url = post.url
        if is_imgur_page(url): url = fetch_imgur_url(url)
        if not url: continue

        # Check image is fetched, then check file is not empty/corrupt.
        fname = re.sub('\?\d*$', '', url.strip('/')).split('/')[-1]
        img_path = os.path.join(img_dir, fname)
        if fetch_img_from_url(url, img_path, verbose):
            # Animated images can't be opened by cv2, which is used by
            # ImageSequence. In future Pillow may be better choice.
            try:
                img = cv2.imread(img_path)
                assert img is not None, 'Image must not be None.'
            except Exception as e:
                if verbose: print(f'IMAGE OPEN ERROR: {fname}')
                os.remove(img_path)
                continue
            posts.append((url, post.permalink))

    # Save csv with image and post urls.
    df = pd.DataFrame(posts, columns=['url', 'post_url'])
    df['post_url'] = 'https://www.reddit.com' + df.post_url
    df.to_csv(os.path.join(out_dir, 'urls.csv'), index=False)
    return df

#Cell
def scrape_subreddit_images(sub, root_dir=None, steps=None, step_size=1_000,
                            after_utc=None, verbose=True, print_freq=50):
    """Scrape reddit images using the pushshift api. This is sometimes
    necessary because PRAW limits us to 1,000 images. Here, we still have to
    collect <=1000 images per api call but we can sort and filter by timestamp
    to accumulate all posts in the subreddit.

    Parameters
    ----------
    sub: str
        Name of subreddit to scrape.
    root_dir: str
        Output directory to save results in. This will contain a subdirectory
        'images' containing images and a text file 'urls.csv' in the root.
        If None is specified, defaults to {sub}_images/.
    steps: int or None
        Number of pushshift api calls to make, where each call contains
        `step_size` posts. Default is None which lets us get all posts in the
        subreddit.
    step_size: int
        Number of posts per api call. Max available is 1,000, but when testing
        we may want to use a smaller number.
    verbose: bool
        If True, print error messages when applicable. This is separate from
        the print_freq functionality, which is concerned with updates rather
        than errors.
    print_freq: int
        Frequency with which to print messages mentioning which image is being
        scraped.
    """
    root_dir = root_dir or sub + '_images'
    img_dir = os.path.join(root_dir, 'test', sub)
    os.makedirs(img_dir, exist_ok=True)

    # Use logger to track last timestamp scraped in case process fails midway.
    logging.basicConfig(handlers=[
            logging.FileHandler(os.path.join(root_dir, 'scrape.log'),
            mode='w')
        ],
        format='%(message)s',
        level=logging.INFO
    )
    logger = logging.getLogger()

    curr_url = push_url = (f'http://api.pushshift.io/reddit/search/submission/'
                           f'?subreddit={sub}&size={step_size}&sort_type='
                           f'created_utc&sort=asc')
    if after_utc: latest = after_utc

    metadata = []
    i = 1
    n_scraped = 0
    while True:
        if i > 1 or after_utc: curr_url = f'{push_url}&after={latest}'
        if verbose: print(f'\nPushshift call #{i}...')

        # Get dict of posts from pushshift.
        r = requests.get(curr_url)
        if r.status_code != 200: break
        data = r.json()['data']
        # When no more posts are found, data is an empty list.
        if not data: break

        # Iterate over each post returned from 1 api call.
        latest = data[-1]['created_utc']
        for post in data:
            img_url = post['url']
            if is_imgur_page(img_url): img_url = fetch_imgur_url(img_url)
            if not img_url: continue

            # Confirm image is fetched, then confirm file is not empty/corrupt.
            fname = re.sub('\?\d*$', '', url.strip('/')).split('/')[-1]
            img_path = os.path.join(img_dir, fname)
            if fetch_img_from_url(img_url, img_path, verbose):
                # Animated images can't be opened by cv2, which is used by
                # ImageSequence. In future Pillow may be better choice.
                try:
                    img = cv2.imread(img_path)
                    assert img is not None, 'Image must not be None.'
                except Exception as e:
                    if verbose: print(f'IMAGE OPEN ERROR: {str(e)} ({fname})')
                    os.remove(img_path)
                    continue

                # Only store for images that were successfully saved.
                metadata.append((post['full_link'], img_url))
                logger.info(post['created_utc'])
                if n_scraped % print_freq == 0:
                    print(f'>>> Scraping image #{n_scraped}: {img_url}')
                n_scraped += 1

        if i == steps: break
        i += 1

    df_kwargs = dict(path_or_buf=os.path.join(root_dir, 'urls.csv'),
                     index=False)
    if after_utc:
        df_kwargs['mode'] = 'a'
        df_kwargs['header'] = False
    df = pd.DataFrame(metadata, columns=['post_url', 'img_url'])
    df.to_csv(**df_kwargs)
    return df

#Cell
def remove_corrupt_images(dir_, mode='move', cv=True, verbose=True):
    """This can be used to remove empty, corrupt, or animated images
    from a directory. Note that some images can be opened by Pillow but
    not cv2, which is what ImageSequence uses to load images.

    Parameters
    ----------
    dir_: str
        Directory of images.
    mode: str
        'move' to move images 1 directory up (safer). 'remove' to remove
        images.
    cv: bool
        True to use cv2, False to use Pillow. Pillow can open animated gifs
        and cv2 can't. Keep in mind spellotape.dl.ImageSequence uses cv2.

    Returns
    -------
    int: Number of images moved or deleted.
    """
    def test_cv2(path):
        img = cv2.imread(path)
        assert img is not None, 'Image must not be None.'

    def test_pillow(path):
        img = Image.open(path)

    open_img = test_cv2 if cv else test_pillow
    changed = 0
    for i, path in enumerate(glob.glob(os.path.join(dir_, '*'))):
        if i % 50 == 0: print(i)
        try:
            open_img(path)
        except Exception as e:
            if verbose: print('error:', e, 'path:', path)
            if mode == 'move':
                path_obj = Path(path)
                os.rename(path,
                          os.path.join(path_obj.parents[1], path_obj.name))
            elif mode == 'remove':
                os.remove(path)
            changed += 1

    return changed


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/06_labeling.ipynb (unless otherwise specified).

__all__ = ['labeler_agreement']

#Cell
import numpy as np

#Cell
def labeler_agreement(df_full, mode='gold', label_col='new_y'):
    """Compute estimate of inter-labeler agreement for a binary classification
    task from a Figure 8 full labeling report. We compute agreement scores for
    each example, then average these scores over all examples to obtain our
    estimate of overall agreement. Remember that due to quality control
    measures, only labelers deemed "reliable" are included - if we were to
    open the task to any labeler, performance would drop.

    Note that this is one of several ways we could estimate human level
    performance, each with its own pros and cons:

    - Take the aggregated report and compute a simple mean of confidence
    scores for all examples.
    - Take the full report and compute a simple mean of trust scores
    (gold set accuracy) for all labelers.
    - Take the full report and compute average agreement scores for each
    example. (This is the method used here.)

    Parameters
    ----------
    df_full: pd.DataFrame
        Full report from Figure 8, where each row contains a single judgment.
    mode: str
        One of ('gold', 'all', 'regular'). "gold" will compute the estimate
        only using labels for the gold standard set (i.e. the test questions).
        "all" will use the whole set for computation. "regular" will use only
        non-test questions.
    label_col: str
        Name of column containing labels to assess.

    Returns
    -------
    float: Value between 0 and 1 measuring the percent of labelers who
        selected the majority class, averaged over all examples in the job.
    """
    if mode == 'gold':
        df = df_full[df_full._golden]
    elif mode == 'regular':
        df = df_full[~df_full._golden]
    elif mode == 'all':
        df = df_full

    def agreement(seq):
        """Binary inter-labeler agreement score.

        Parameters
        ----------
        seq: list, np.array, pd.Series
            A sequence of binary labels for a single example.

        Returns
        -------
        float: Value between .5 and 1 measuring inter-labeler agreement.
        """
        avg = np.mean(seq)
        return max(avg, 1-avg)

    return df.groupby('_unit_id')[label_col].apply(agreement).mean()


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/02_nlp.ipynb (unless otherwise specified).

__all__ = ['nlp', 'tokenize', 'lexical_density']

#Cell
import os
import spacy

#Cell
nlp = spacy.load('en_core_web_sm')

#Cell
def tokenize(raw, lower=True):
    """Tokenize a string of text.

    Parameters
    -----------
    raw: str
        The text to tokenize.
    lower: bool
        If True, lowercase the input text.

    Returns
    --------
    list[str]
    """
    if lower: raw = raw.lower()
    return [t.text for t in nlp(raw, disable=['tagger', 'parser', 'ner'])]

#Cell
def lexical_density(text, lower=True):
    """Compute lexical density of a piece of text.

    https://en.wikipedia.org/wiki/Lexical_density

    Parameters
    -----------
    text: str or list[str]
    lower: bool (see `tokenize` docs)

    Returns
    --------
    float: Number between 0 and 1, where larger values indicate a higher
        lexical density.
    """
    if isinstance(text, str):
        text = tokenize(text, lower)
    if not text:
        return 0.0
    return len(set(text)) / len(text)


#AUTOGENERATED! DO NOT EDIT! File to edit: dev/04_dl.ipynb (unless otherwise specified).

__all__ = ['IMAGENET_STATS', 'gpu_found', 'find_layer_groups', 'inverse_sigmoid', 'mish', 'normalize', 'denormalize',
           'show_batch', 'show_image_variants', 'ImageSequence', 'CosineLRScheduler', 'BatchHistory',
           'PerformanceThreshold', 'plot_history', 'show_predictions', 'hybrid_quantize']

#Cell
import bz2
import cv2
from itertools import chain, zip_longest
import matplotlib.pyplot as plt
import numpy as np
from operator import gt, lt
import os
import pandas as pd
import pickle
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.layers import Dense, Input
from tensorflow.keras.utils import Sequence
from tensorflow.python.client import device_lib
import warnings

from htools import AutoInit, timeboxed, auto_repr

#Cell
IMAGENET_STATS = np.array([[0.485, 0.456, 0.406],
                           [0.229, 0.224, 0.225]])

#Cell
def gpu_found():
    """Check if any tensorflow can find any gpu's on device.

    Returns
    -------
    bool
    """
    return any('GPU' in str(dev) for dev in device_lib.list_local_devices())

#Cell
def find_layer_groups(model, percent=.5, skip_head=True, plot=True):
    """Tool to help determine model layer groups.

    Parameters
    ----------
    model: tf.keras.Model
    percent: float
        Percent of model weights to find the split point for. For example,
        .3 will find the layer such that 30% of the model weights lie in or
        before that layer.
    skip_head: bool
        Specify whether to exclude the last layer from calculations. We often
        choose to treat the last layer as its own group for transfer learning
        (group 3), so if we want group 1 and 2 to be equally sized in this
        scenario, we must use skip_head=True.
    plot: bool
        If True, plot the number of weights vs. layer number.

    Returns
    -------
    tuple(str, float): Name of layer, cumulative % of model weights contained
        in that layer and its preceding layers.
    """
    name2size = {k: np.prod(v) for k, v in model.layer_dims().items()}
    name2cum = dict(zip(name2size.keys(), np.cumsum(list(name2size.values()))))

    if skip_head:
        total = sorted(name2cum.values())[-2]
    else:
        total = max(name2cum.values())
    cutoff = total * percent

    if plot:
        plt.plot(list(name2cum.values()))
        plt.axhline(cutoff, c='red', ls='--')
        plt.show()

    # Return name of the layer where the cumulative total of weights in the
    # model first surpasses the specified percent. This is approximately where
    # group 2 could start.
    for k, v in name2cum.items():
        if v > cutoff:
            return k, v / total

#Cell
def inverse_sigmoid(y):
    """Use to determine the bias initializer for the final linear layer of
    model.

    Parameters
    ----------
    x: float
        Value between 0 and 1 (e.g. the proportion of the training data that
        are postives).

    Returns
    -------
    float: Inverse sigmoid of input.
        I.E. if y=sigmoid(x), inverse_sigmoid(y)=x.
    """
    return np.log(y / (1-y))

#Cell
def mish(x):
    """Mish: A Self Regularized Non-Monotonic Neural Activation Function

    https://arxiv.org/pdf/1908.08681v1.pdf

    Parameters
    ----------
    x: tf.Tensor[float]
        Input tensor.

    Returns
    -------
    tf.Tensor[float]: Tensor of same shape as input x.
    """
    return x * K.tanh(K.softplus(x))

#Cell
def normalize(img, stats=None):
    """Normalize an image using mean and standard deviation.
    When using pretrained models, the stats should come from the dataset used
    for training, not our data.

    Parameters
    ----------
    img: np.array
        Image or batch of images with channel dimension last.
    stats: np.array
        Channel-wise image stats, where stats[0] contains means and stats[1]
        contains standard deviations. If None (the default), imagenet stats
        will be used.

    Returns
    -------
    """
    # Truth value of array is ambiguous so check explicitly.
    if stats is None:
        stats = IMAGENET_STATS
    return (img - stats[0]) / stats[1]

#Cell
def denormalize(img, stats=None):
    """Reverse normalization process done using mean and standard
    deviation. When using pretrained models, the stats should come from the
    dataset used for training, not our data.

    Parameters
    ----------
    img: np.array
        Image or batch of images with channel dimension last.
    stats: np.array
        Channel-wise image stats, where stats[0] contains means and stats[1]
        contains standard deviations. If None (the default), imagenet stats
        will be used.

    Returns
    -------
    """
    if stats is None:
        stats = IMAGENET_STATS
    return img * stats[1] + stats[0]

#Cell
def show_batch(data_gen, denorm=False, size=(8, 8), max_images=36, **kwargs):
    """
    Parameters
    ----------
    data_gen: DirectoryIterator
        Object returned by ImageDataGenerator.flow_from_directory().
    denorm: bool
        Specify whether denormalization is needed.
    size: tuple[int]
        Figure size.
    max_images: int
        Max number of images to display. When batch sizes are large,
        individual images may be too small to see without either
        increasing size or decreasing max_images.
    kwargs: stats
        If specified, will be passed to denormalize function.
    """
    bs = min(data_gen.batch_size, max_images)
    ncols = int(np.ceil(np.sqrt(bs)))
    i2class = {v: k for k, v in data_gen.class_indices.items()}
    x, y = map(lambda x: x[:bs], next(data_gen))

    # Some generators normalized images w/ imagenet stats.
    if denorm: x = denormalize(x, **kwargs)

    # Create square grid of images. Each image will be labeled with class name.
    fig, ax = plt.subplots(ncols, ncols, figsize=size)
    if ncols == 1: ax = np.array(ax)

    # Must turn off axis for each subplot separately.
    for x, y, ax_ in zip_longest(x, y, ax.flatten()):
        try:
            ax_.imshow(x)
            ax_.set_title(i2class[y])
        except TypeError:
            pass
        ax_.axis('off')

    plt.tight_layout()
    plt.show()

#Cell
def show_image_variants(data_gen, n=4, i=1):
    """Plot n variants of an image from a data generator to get
    a sense of how image augmentation is working. Images are denormalized
    with ImageNet stats before being displayed.

    Parameters
    ----------
    data_gen: iterable
        ImageDataGenerator.flow_from_directory() where each batch is
        the size of the whole dataset. For this reason, it will be
        much faster if constructed from a directory with few images.
    n: int
        Number of variation to display.
    i: int
        Selects i'th image in batch.
    """
    # Plot the images given by the iterator.
    imgs = (denormalize(next(data_gen)[0][i]) for _ in range(n))
    fig, ax = plt.subplots(nrows=i, ncols=n, figsize=(18, 18))
    for ax_i, img in zip(ax, imgs):
        ax_i.imshow(img)
        ax_i.axis('off')
    plt.show()

#Cell
class ImageSequence(Sequence):

    def __init__(self, dir_,  bs, pos_class, image_shape=(224, 224, 3),
                 transforms=None, rescale=1/255, shuffle=True, max_len=None,
                 save_path=None, imagenet_norm=True, seed=1):
        """
        Parameters
        ----------
        dir_: str
            Parent directory. Must contain sub-directories for each class.
        bs: int
            Size of each mini batch.
        pos_class: str
            Name of class to treat as positive (label of 1). Must match the
            name of the directory containing these images.
        image_shape: tuple(int)
            Shape of image with channels last (H x W x C).
        transforms: albumentations.core.composition.Compose
            Compose object from albumentations package. This must be callable
            such that transforms(image=x)['image'] returns the transformed
            image x.
        rescale: float
            Value to multiply image rgb values by. For a standard image
            dataset, the default 1/255 will scale values to lie between 0 and
            1.
        shuffle: bool
            Whether to shuffle the data after each epoch. This will result in
            different mini batches, which may have a regularizing effect.
        max_len: int (optional)
            If specified, will limit the number of samples in the dataset. This
            makes it easier to experiment with subsets if we have a huge number
            of files.
        save_path: str
            Path to save filenames to. Should end with a pickle file.
        imagenet_norm: bool
            If True, normalize dataset with imagenet mean and standard
            deviation.
        seed: int
            Random seed used when shuffling data, if shuffle=True. User may
            pass seed=None to avoid setting a seed.
        """
        if seed is not None:
            np.random.seed(seed)

        self.dir_ = dir_
        self.image_shape = image_shape
        self.bs = bs
        self.pos_class = pos_class
        self.transforms = transforms
        self.rescale = rescale
        self.shuffle = shuffle
        self.save_path = save_path
        self.norm = imagenet_norm
        if self.norm:
            self.means = np.array([0.485, 0.456, 0.406])
            self.stds = np.array([0.229, 0.224, 0.225])

        # Compile list of file paths. Labels will be generated on the fly.
        self.files = self.get_files(max_len)
        self.i2label = self._get_labels()
        self.n_samples = len(self.files)
        self.on_epoch_end()

        if save_path:
            self.save()

    def on_epoch_end(self):
        if self.shuffle:
            np.random.shuffle(self.files)

    def load_img(self, path):
        img = cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB)
        return cv2.resize(img, self.image_shape[:-1])

    def normalize(self, img):
        """Normalize an image using mean and standard deviation. When using
        pretrained models, the stats should come from the dataset used
        for training, not our data.

        Parameters
        ----------
        img: np.array
            Image or batch of images with channel dimension last.
        stats: np.array
            Channel-wise image stats, where stats[0] contains means and
            stats[1] contains standard deviations. If None (the default),
            imagenet stats will be used.

        Returns
        -------
        np.array
        """
        return (img - self.means) / self.stds

    def denormalize(self, img):
        """Reverse normalization process done using mean and standard
        deviation. When using pretrained models, the stats should come from the
        dataset used for training, not our data.

        Parameters
        ----------
        img: np.array
            Image or batch of images with channel dimension last.
        stats: np.array
            Channel-wise image stats, where stats[0] contains means and
            stats[1] contains standard deviations. If None (the default),
            imagenet stats will be used.

        Returns
        -------
        np.array
        """
        return img * self.stds + self.means

    def get_files(self, n):
        """Get list of filenames, discarding any beyond the desired dataset
        size. Shuffling occurs here regardless of self.shuffle choice because
        we want a representative dataset, not all 1 class.
        """
        files = list(self._get_files(self.dir_))
        np.random.shuffle(files)
        return files[:n]

    @staticmethod
    def _get_files(dir_):
        """Get files contained in all subdirectories of a parent directory.

        Parameters
        ----------
        dir_: str
            Name of parent directory. It should contain a subdirectory for each
            class.

        Returns
        -------
        itertools.chain: Generator with all files in subdirectories.
        """
        return chain.from_iterable([
            [f.path for f in os.scandir(arg.path)]
            for arg in os.scandir(dir_)
            if arg.is_dir() and not arg.name.startswith('.')
        ])

    def _get_labels(self):
        """Called automatically on instantiation to create a class
        number to class name mapping in the form of a list.

        Returns
        -------
        List[str]: List of labels, where the positive class is last
            (i.e. in binary classification for classes SFW and NSFW
             where NSFW is positive, we return ['SFW', 'NSFW']).
        """
        labels = [arg.name for arg in os.scandir(self.dir_) if arg.is_dir()]
        return sorted(labels, key=lambda x: x == self.pos_class)

    def show_batch(self, i=None, denorm=True, size=(8, 8)):
        """
        Parameters
        ----------
        i: int
            Batch number to display. If not specified, a random batch is
            chosen.
        denorm: bool
            Specify whether denormalization is needed.
        size: tuple[int]
            Figure size.
        """
        if i is None:
            i = np.random.randint(0, len(self))
        x, y = self[i]
        ncols = int(np.ceil(np.sqrt(self.bs)))

        # Ensure images are displayed correctly and without warning.
        if denorm:
            x = self.denormalize(x)
        x = x.round(5)

        # Plot each image and label with class name.
        fig, ax = plt.subplots(ncols, ncols, figsize=size)
        if ncols == 1:
            ax = np.array(ax)

        # Must turn off axis for each subplot separately.
        for i, (x_, y_, ax_) in enumerate(zip_longest(x, y, ax.flatten())):
            try:
                ax_.imshow(x_)
                ax_.set_title(self.i2label[int(y_)])
            except TypeError:
                pass
            ax_.axis('off')

        plt.tight_layout()
        plt.show()

    def save(self):
        """Pickle and zip generator for later use. Results should be
        reproducible anyway when setting a random seed, but this gives us extra
        insurance.
        """
        print(f'Saving ImageGenerator to {self.save_path}.')
        with bz2.BZ2File(self.save_path, 'w') as f:
            pickle.dump(self, f)

    @staticmethod
    def from_zipfile(path):
        """Load a previously saved ImageSequence.
        """
        with bz2.BZ2File(path, 'r') as f:
            seq = pickle.load(f)
        return seq

    def __len__(self):
        """Returns number of batches per epoch, NOT the number of total images.
        """
        return int(np.ceil(self.n_samples / self.bs))

    def _get_one_xy(self, i):
        """Load a single (image, label) pair.

        Parameters
        ----------
        i: int
            Integer used to index into list of file names.

        Returns
        -------
        tuple[np.array, int]: x, y
        """
        path = self.files[i]
        y = path.split('/')[-2] == self.pos_class
        x = self.load_img(path)
        if self.transforms:
            x = self.transforms(image=x)['image']
        if self.norm:
            x = self.normalize(x * self.rescale)
        return x, y

    def __getitem__(self, i):
        """Retrieve 1 batch.

        Returns
        -------
        np.array, np.array: First item is X, a mini batch of images
            with shape: (batch size, height, width, channels).
            Second item is y, the corresponding labels where 1 indicates NSFW
            and 0 indicates SFW, with shape: (batch size, ).
        """
        # Make sure that we don't select index larger than max idx.
        idx = range(i * self.bs, min(len(self.files), (i+1) * self.bs))
        xy_pairs = [self._get_one_xy(i) for i in idx]
        x, y = map(lambda arg: np.array(arg, dtype=np.float32), zip(*xy_pairs))
        return x, y

    def __eq__(self, obj):
        """Check if two ImageSequence objects are equivalent. Useful for
        debugging and when trying to reproduce results.
        """
        if isinstance(obj, ImageSequence):
            return all(v == getattr(obj, k)
                       for k, v in self._comparable_attrs().items())
        return False

    def __ne__(self, obj):
        """Check if two objects are not equal. Calls __eq__ implicitly.
        """
        return not self == obj

    def _comparable_attrs(self):
        """Helper method for checking equality.
        """
        return {k: v for k, v in self.__dict__.items()
                if k not in ('means', 'stds')}

    def __repr__(self):
        return (f'ImageSequence(dir_={self.dir_}, bs={self.bs}, '
                f'image_shape={self.image_shape}), shuffle={self.shuffle})')

#Cell
class CosineLRScheduler(Callback):
    """Learning rate scheduler that makes updates each batch. Built-in
    Keras version only updates once per epoch.
    """

    def __init__(self, max_lr, epochs, iters_per_epoch, warm=0.3,
                 restarts=False, cycle_len=5, cycle_decay=0.0, min_lr=None,
                 verbose=False):
        """
        Parameters
        ----------
        max_lr: float
            Maximum learning rate to use during training.
        epochs: int
            Number of epochs to train for.
        iters_per_epoch: int
            Number of batches in each epoch.
        warm: float
            Percent of training run (or cycle length) devoted to the increasing
            portion of the schedule. Default 0.3.
        restarts: bool
            Specifies whether to use restarts, i.e. use a cyclical LR.
            True: Version of cosine annealing with restarts. In one
                  cycle, LR starts high and gradually decreases.
                  At the start of the next cycle, it is
                  immediately increased again.
            False: Version of cosine annealing where LR increases
                   for first 30% of training, then decreases for
                   remaining 70%.
        cycle_len: int
            Number of epochs contained in a single cycle. Only used
            when scheduler uses restarts.
        cycle_decay: float
            Scalar to decay the learning rate at the end of each cycle.
            This is only used with restarts, since the regular cosine
            annealing already decays the LR over time.
            E.g. 1.0 will use no decay.
            0.9 means that cycle 2 LRs = cycle 1 LRs * 0.9,
            cycle 3 LRs = cycle 1 LRs * .81,
            etc.
        min_lr: float
            Minimum learning rate. If None is specified, it will be set
            to max_lr / 25.
        """
        super().__init__()
        self.max_lr = max_lr
        self.min_lr = min_lr or max_lr / 25
        self.epochs = epochs
        self.iters_per_epoch = iters_per_epoch
        self.iters = epochs * iters_per_epoch
        self.warm = warm

        if restarts and cycle_len > self.iters_per_epoch:
            warnings.warn('Epoch consist of less than 1 full cycle.')

        self.cycle_len = cycle_len
        self.cycle_decay = cycle_decay
        self.restarts = restarts
        self.lrs = self._schedule(restarts)
        self.curr_iter = 0
        self.verbose = verbose

    def on_batch_begin(self, batch, logs=None):
        """Update learning rate based on current iteration."""
        if self.verbose:
            print(f'\nUpdate LR from {self.model.optimizer.lr.numpy():.3E}',
                  end='')

        # Continue using min_lr if training past specified max iters.
        try:
            lr = self.lrs[self.curr_iter]
        except IndexError:
            lr = self.lrs[-1]
        K.set_value(self.model.optimizer.lr, lr)

        self.curr_iter += 1
        if self.verbose:
            print(f' to {self.model.optimizer.lr.numpy():.3E}')

    @staticmethod
    def _cosine_anneal(iters, lr1, lr2):
        """Helper function for _cosine_tri_lr().

        Parameters
        ----------
        iters: int
            Number of iterations in segment.
        lr1: float
            Learning rate at start of segment.
        lr2: float
            Learning rate at end of segment.

        Returns
        -------
        np.array
        """
        i = np.arange(iters)
        return lr2 + (lr1 - lr2)*(1 + np.cos(np.pi * i/iters))/2

    def _cosine_schedule(self):
        """Cosine annealing scheduler. Computes learning rates for each
        iteration.

        Returns
        -------
        np.array
        """
        seg1 = self._cosine_anneal(int(self.warm * self.iters),
                                   self.min_lr,
                                   self.max_lr)
        seg2 = self._cosine_anneal(int(np.ceil((1 - self.warm) * self.iters)),
                                   self.max_lr,
                                   self.min_lr)
        return np.concatenate((seg1, seg2))

    def _cosine_restarts_schedule(self):
        """Cosine annealing with restarts."""
        cycles = int(np.ceil(self.iters /
                             (self.cycle_len * self.iters_per_epoch)))
        cycle_iters = self.cycle_len * self.iters_per_epoch
        lrs = [self._cosine_anneal(cycle_iters, self.max_lr, self.min_lr)
               / (1 + self.cycle_decay * i) for i in range(cycles)]
        return np.concatenate(lrs)

    def _schedule(self, restarts):
        """Wrapper to schedule learning rates depending on chosen method.

        Parameters
        ----------
        restarts: bool
            If True, use schedule with restarts. If False, use regular
            cosine annealing that spans whole duration of training.

        Returns
        -------
        np.array: LR for each iteration (i.e. output[i] is the LR to use
            at iteration i).
        """
        if restarts:
            return self._cosine_restarts_schedule()
        return self._cosine_schedule()

    def plot_lrs(self, path=None):
        """Display learning rate by iteration.

        Note: If the plot is not as smooth as expected, this likely
        means that there are very few iterations per epoch
        (i.e. the batch size is very large, at least in relative terms).
        """
        plt.plot(self.lrs)
        plt.xlabel('Iteration')
        plt.ylabel('Learning Rate')
        plt.title('Learning Rate Schedule')
        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.show()

    def reset(self):
        """Reset the current iteration to zero. Note that if the learning rate
        has been decayed, it will now be back at the initial value.
        """
        self.curr_iter = 0

    def __repr__(self):
        return (f'CosineLRScheduler(max_lr={self.max_lr}, '
                f'epochs={self.epochs}, '
                f'iters_per_epoch={self.iters_per_epoch}, '
                f'restarts={self.restarts}, '
                f'cycle_len={self.cycle_len}, '
                f'min_lr={self.min_lr}, '
                f'verbose={self.verbose})')

#Cell
class BatchHistory(Callback):
    """Callback to store training loss at each mini batch. Built in
    history callback only stores epoch losses. Validation set is only
    evaluated at the end of each epoch so it's not included here.
    """

    def __init__(self, path):
        """Parameters
        -------------
        path: str
            Path to save csv of batch losses to.
        """
        super().__init__()
        self.path = path
        self.losses = []

    def on_batch_end(self, batch, logs=None):
        self.losses.append(logs.get('loss'))

    def on_train_end(self, logs=None):
        with open(self.path, 'a') as f:
            f.write('\n'.join(str(loss) for loss in self.losses) + '\n')

    def plot_batch_losses(self, path=None, smooth=25):
        df = pd.DataFrame(self.losses, columns=['loss'])
        df['smoothed'] = df.loss.ewm(span=smooth).mean()

        # Plot losses and smoothed losses.
        plt.plot(df.loss, lw=.5, c='blue', alpha=.25)
        plt.plot(df.smoothed, lw=.5, c='blue', alpha=1)
        plt.xlabel('Iteration')
        plt.ylabel('Training Loss')
        plt.title('Training Loss by Mini Batch.')
        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.show()

#Cell
class PerformanceThreshold(Callback):
    """Callback that will stop training if performance falls short of
    some user-specified threshold. This can help avoid wasting time on
    runs that have clearly gone wrong - e.g. learning rate is too large
    and validation BCE loss has ballooned to >2.0 on a balanced dataset.
    """

    def __init__(self, metric='val_loss', threshold=1.0, goal='min'):
        """
        Parameters
        ----------
        metric: str
            Name of metric to monitor.
        threshold: float
            Performance threshold. If the quantity being monitored is
            something we want to minimize, this is the highest
            acceptable value to continue training. If it's a metric to
            maximize like accuracy, this is the minimum acceptable
            value.
        goal: str
            Goal of the quantity being monitored (i.e. are we trying to
            minimize metric or maximize it?). One of ('max', 'min').
        """
        super().__init__()
        assert goal in ('max', 'min')

        self.metric = metric
        self.threshold = threshold
        if goal == 'max':
            self.op = gt
        elif goal == 'min':
            self.op = lt

    def on_epoch_end(self, batch, logs=None):
        curr_val = logs.get(self.metric)
        if curr_val is not None and self.op(self.threshold, curr_val):
            print(f'{self.metric} of {curr_val:.3e} did not satisfy '
                  f'performance threshold of {self.threshold}. '
                   'Stopping training.')
            self.model.stop_training = True

#Cell
def plot_history(*args, vlines=False, metrics=None, path=None):
    """Plot metrics from one or more training run.

    Parameters
    ----------
    args: tensorflow.keras.callbacks.history
        One or more history objects returned from training. If multiple history
        objects are passed in, their stats will be concatenated.
    vlines: bool
        If True, add vertical lines to the plots showing where each training
        run ended.
    metrics: list[str]
        Names of metrics to plot. If excluded, all metrics will be plotted
    """
    if not metrics:
        metrics = {metric.replace('val_', '')
                   for metric in args[0].history.keys()}

    # Concatenate metric lists if multiple history objects passed in.
    hist = {m: list(chain.from_iterable([arg.history[m] for arg in args]))
            for m in args[0].history.keys()}
    train_lens = np.cumsum([len(arg.history['loss']) for arg in args])

    # Create subplot for each metric.
    length = len(metrics)
    fig, ax = plt.subplots(length, 1, figsize=(8, length*2.5))
    for i, m in enumerate(metrics):
        ax[i].set_title(m.title())
        ax[i].plot(hist[m], label=m)

        # Plot validation metric if present.
        val_metric = 'val_' + m
        try:
            ax[i].plot(hist[val_metric], label=val_metric)
        except KeyError:
            pass

        ax[i].legend()
        if not vlines:
            continue

        # Show where each training run ended.
        for length in train_lens[:-1]:
            ax[i].axvline(length, lw=0.8, alpha=0.9, ls='--', c='grey')

    plt.subplots_adjust(hspace=.4)
    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.show()

#Cell
def show_predictions(df, dg, mode, n=9, sort=True, most_wrong=True,
                     figsize=(8, 8), threshold=None, y_true='y',
                     y_pred='y_pred', y_proba='y_proba'):
    """Display images based on the model's predictions. You can view false
    positives, false negatives, or a random subset of all images.

    Parameters
    ----------
    df: pd.DataFrame
        Labels and predictions df produced by evaluate_model.py script.
    dg: ImageSequence
        Same data generator that was used to produce the predictions in df.
        Because shuffle=False in evaluation script, row i in df corresponds to
        file i in dg.
    mode: str
        'fp' to show false positives, 'fn' to show false negatives, 'any'
        to show any predictions (both correct and incorrect).
    n: int
        Number of images to display.
    sort: bool
        Whether to sort predictions by distance from true label before
        displaying. I.e. if True, will show top 5 "most wrong" predictions.
        If False, shows first 5 wrong predictions.
    most_wrong: bool
        If True, show the predictions where the model was most wrong. If False,
        show the "least wrong" errors (model was wrong but nearly predicted
        the correct class). To get a random selection of images, set
        sort=False.
    threshold: float or None
        If specified, will re-compute hard predictions using this as the
        decision threshold. If None, df must already have column of hard
        predictions.
    hard_col: str
        Name of column in df containing hard predictions.
    soft_col: str
        Name of column in df containing soft predictions.
    """
    if threshold: df = update_hard_preds(df, threshold, soft_col)

    if mode == 'fp':
        df = df[df[y_true] < df[y_pred]]
    elif mode == 'fn':
        df = df[df[y_true] > df[y_pred]]
    elif mode != 'any':
        raise ValueError('Parameter `mode` must be "fp", "fn" or "any".')

    # Sort after filtering hard labels since we may use very high thresholds
    # at times, which could lead to correctly predicted labels that have large
    #  distances between predicted prob and true label.
    ascending = not most_wrong
    if sort:
        df = df.lambda_sort(lambda x: abs(x[y_true] - x[y_proba]),
                            ascending=ascending)
    else:
        df = df.sample(frac=1, replace=False)

    ncols = int(np.ceil(np.sqrt(n)))
    fig, ax = plt.subplots(ncols, ncols, figsize=figsize)
    for idx, ax_ in zip_longest(df.index[:n], ax.flatten()):
        try:
            path = dg.files[idx]
            img = dg.load_img(path)
            ax_.imshow(img)
            title = (f'{path.split("/")[-1]}\n '
                     f'{df.loc[idx, y_proba]:.4f}')
            ax_.set_title(title[-30:], size=10)
        except:
            pass
        ax_.axis('off')

    plt.tight_layout()
    plt.show()

#Cell
def hybrid_quantize(model, path=None):
    """Perform post-training hybrid quantization. Weights will be
    changed from float32 to int8, while biases and activation remain as
    float32. TF claims that a 75% reduction in model size is typical, with
    small effects on accuracy.

    Parameters
    ----------
    model: tf.keras.models.Model
    path: str
        Optionally specify an output path to save the quantized weights to.

    Returns
    -------
    bytes or None: If no path is provided, returns quantized weights. If
        a path is provided, save model and return None.
    """
    # Reduced float quantization.
    converter = tf.lite.TFLiteConverter.from_keras_model(model)
    converter.optimizations = [tf.lite.Optimize.OPTIMIZE_FOR_SIZE]
    quantized_bytes = converter.convert()

    # Save quantized weights if requested.
    if path:
        with open(path, 'wb') as f:
            f.write(quantized_bytes)
            print(f'Quantized bytes saved to {path}.')
            return
    return quantized_bytes


#AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"IMG_EXT": "00_utils.ipynb",
         "cd_root": "00_utils.ipynb",
         "extension_counts": "00_utils.ipynb",
         "stop_instance": "00_utils.ipynb",
         "snapshot_protobufs": "01_experiments.ipynb",
         "fetch_experiments_data": "01_experiments.ipynb",
         "nlp": "02_nlp.ipynb",
         "tokenize": "02_nlp.ipynb",
         "lexical_density": "02_nlp.ipynb",
         "Debugger": "03_ml.ipynb",
         "build_pipeline": "03_ml.ipynb",
         "parse_gs_results": "03_ml.ipynb",
         "IMAGENET_STATS": "04_dl.ipynb",
         "gpu_found": "04_dl.ipynb",
         "find_layer_groups": "04_dl.ipynb",
         "inverse_sigmoid": "04_dl.ipynb",
         "mish": "04_dl.ipynb",
         "normalize": "04_dl.ipynb",
         "denormalize": "04_dl.ipynb",
         "show_batch": "04_dl.ipynb",
         "show_image_variants": "04_dl.ipynb",
         "ImageSequence": "04_dl.ipynb",
         "CosineLRScheduler": "04_dl.ipynb",
         "BatchHistory": "04_dl.ipynb",
         "PerformanceThreshold": "04_dl.ipynb",
         "plot_history": "04_dl.ipynb",
         "show_predictions": "04_dl.ipynb",
         "hybrid_quantize": "04_dl.ipynb",
         "ClickLiteralOption": "05_cli.ipynb",
         "click.command": "05_cli.ipynb",
         "click.option": "05_cli.ipynb",
         "labeler_agreement": "06_labeling.ipynb",
         "fetch_img_from_url": "07_scraping.ipynb",
         "load_praw_credentials": "07_scraping.ipynb",
         "is_imgur_page": "07_scraping.ipynb",
         "fetch_imgur_url": "07_scraping.ipynb",
         "scrape_subreddit_images_1k": "07_scraping.ipynb",
         "scrape_subreddit_images": "07_scraping.ipynb",
         "remove_corrupt_images": "07_scraping.ipynb"}

modules = ["utils.py",
           "experiments.py",
           "nlp.py",
           "ml.py",
           "dl.py",
           "cli.py",
           "labeling.py",
           "scraping.py"]

git_url = "https://github.com/hdmamin/spellotape/tree/master/"

def custom_doc_links(name): return None